import urllib, urllib2, md5, json, re, os, shutil, argparse
from os import path
from zipfile import ZipFile

parser = argparse.ArgumentParser()
parser.add_argument('--pack', help='name of the pack, one word', required=True)
args = parser.parse_args()

base_pack_url = u'http://rianon.info/minecraft/' + args.pack + u'/'
config_fn = u'config.json'

runtime_dir = path.dirname(path.abspath(unicode(__file__, 'utf8')))
config_md5_file = path.join(runtime_dir, 'config.md5')
download_dir = path.join(runtime_dir, 'downloads')
mc_dir = path.abspath(path.join(runtime_dir, '../.minecraft'))


def verify_dir_exists(dirname):
    if not path.exists(dirname):
        os.makedirs(dirname)


def http_get(fileurl):
    try:
        return urllib2.urlopen(base_pack_url + urllib.pathname2url(fileurl))
    except:
        return None


def file_md5(filename):
    return md5.new(file(filename, 'rb').read()).hexdigest()


def read_client_hash():
    return path.exists(config_md5_file) and file(config_md5_file, 'r').read() or ''


def save_client_hash(md5):
    file(config_md5_file, 'w').write(md5)


def extract_zip_without_metainf(fn, target_dir):
    if not path.exists(fn):
        exit('Error: file not found: {0}'.format(fn))

    with ZipFile(fn, 'r') as src:
        for name in src.namelist():
            if not re.match('^META\-INF', name):
                src.extract(name, target_dir)


def unpack_reference_bin():
    print 'updating minecraft binaries'
    mc_versions = path.join(mc_dir, 'versions')
    mc_libraries = path.join(mc_dir, 'libraries')
    if path.exists(mc_versions):
        shutil.rmtree(mc_versions)
    if path.exists(mc_libraries):
        shutil.rmtree(mc_libraries)
    extract_zip_without_metainf(path.join(download_dir, 'reference_bin.zip'), mc_dir)
    extract_zip_without_metainf(path.join(download_dir, 'reference_libs.zip'), mc_dir)


def unpack_reference_configs():
    print 'updating minecraft configs'
    # do not delete existing configs
    extract_zip_without_metainf(path.join(download_dir, 'reference_config.zip'), mc_dir)


def update_mc_resource(res, cfg):
    if cfg is None:
        exit('Error: received empty cfg for resource {0}'.format(res))

    print 'updating', res
    cache_dir = path.join(download_dir, res)
    dest_dir = path.join(mc_dir, res)
    verify_dir_exists(dest_dir)

    for fn in cfg:
        src_file = path.join(cache_dir, fn)
        dest_file = path.join(dest_dir, fn)

        if path.exists(dest_file) and file_md5(dest_file) == cfg[fn]:
            continue
        
        # remove if exist different version
        if path.exists(dest_file):
            os.remove(dest_file)

        print 'copying file', path.basename(src_file)
        shutil.copy(src_file, dest_file)


def update_file_if_changed(fn, save_to, remote_dir, remote_md5):

    verify_dir_exists(save_to)

    abspath = path.abspath(path.join(save_to, fn))
    fileurl = path.join(remote_dir, fn)

    if path.exists(abspath):
        if file_md5(abspath) == remote_md5:
            return False
        os.remove(abspath)

    print 'downloading <<', fileurl
    response = http_get(fileurl)
    if response is None:
        exit('Error: file not found on server, please contact pack admin')
    
    with open(abspath, 'wb') as f:
        f.write(response.read())
        f.close()

    if file_md5(abspath) != remote_md5:
        exit('Error: downloaded file is corrupted, md5 does not match')

    return True


def update_resource_cache(res, cfg):
    save_to = path.join(download_dir, res)
    for fn in cfg:
        update_file_if_changed(fn, save_to, res, cfg[fn])


# remove files not listed in cfg from cache_dir, and their copies from installed_dir
def cleanup_dir(cache_dir, installed_dir, cfg):
    if not path.exists(cache_dir):
        return
    print 'cleaning', path.basename(cache_dir)
    for fn in os.listdir(cache_dir):
        if fn not in cfg:
            print 'removing deprecated file:', fn
            if installed_dir is not None:
                installed_fn = path.join(installed_dir, fn)
                if path.exists(installed_fn):
                    os.remove(installed_fn)
            os.remove(path.join(cache_dir,fn))


# remove old files from downloads and .minecraft/*
def cleanup_all_resources(cfg):
    for res in cfg:
        if isinstance(cfg[res], dict):
            cleanup_dir(path.join(download_dir, res), path.join(mc_dir, res), cfg[res])


def update_pack(cfg, likenew):
    for res in cfg:
        if isinstance(cfg[res], dict):
            update_resource_cache(res, cfg[res])
        elif not res.startswith('runtime'):
            # set likenew to True if reference_bin.zip or reference_config.zip were updated
            likenew |= update_file_if_changed(res, download_dir, '', cfg[res])

    # dont forget to clean up old files
    cleanup_all_resources(cfg)

    if likenew:
        unpack_reference_bin()
        unpack_reference_configs()

    # copy mods, ...
    for res in ['mods', 'shaderpacks']:
        update_mc_resource(res, cfg[res])


# main entry
if __name__ == '__main__':

    likenew = not path.exists(mc_dir)

    print 'downloading <<', config_fn
    config_response = http_get(config_fn)

    if config_response is None:
        exit('config file not found on server, please contact pack admin')

    raw_config = config_response.read()
    config_md5 = md5.new(raw_config).hexdigest()

    if read_client_hash() == config_md5 and not likenew:
        exit('client is up to date')

    print 'updating pack contents'
    update_pack(json.loads(raw_config), likenew)
    
    print '...done, saving config hash'
    save_client_hash(config_md5)
